# -----------------------------------------------------------------------------
# Read and write segmentation data in hdf5 format.
#
# Example layout:
#
# format = "segger"
# format_version = 1
#
# name = "somedata segmentation"
# mask = <3-d array of region indices>
# region_ids = <array or region ids numbers, length N>
# region_colors = <array of rgba, N by 4>
# ref_points = <array of region reference points, N by 3>
# parent_ids = <array each regions parent (0 = no parent), length N>
# smoothing_levels = <array of float, length N>
#
# map_path = "/Users/smith/somedata.mrc"
# map_size = (512, 512, 200)
# map_level = 1.245
# ijk_to_xyz_transform = <3x4 matrix>
#
# skeleton = 'string encoding chimera marker file'
#
# The file is saved with the Python PyTables modules which includes
# additional attributes "VERSION", "CLASS", "TITLE", "PYTABLES_FORMAT_VERSION".
#
# Tests with alternate data storage with every region being a separate HDF
# node and every contact being a separate HDF node gave extremely slow
# read/write speed.
#
def write_segmentation(seg, path = None):

  if path is None:
    show_save_dialog(seg)
    return

  import tables
  h5file = tables.openFile(path, mode = 'w')

  root = h5file.root
  a = root._v_attrs
  a.format = 'segger'
  a.format_version = 2
  a.name = seg.name

  m = seg.mask
  atom = tables.Atom.from_dtype(m.dtype)
  filters = tables.Filters(complevel=5, complib='zlib')
  ma = h5file.createCArray(root, 'mask', atom, m.shape, filters = filters)
  ma[:] = m

  from numpy import array, int32, float32
  rlist = seg.id_to_region.values()
  rlist.sort(lambda r1,r2: cmp(r1.rid,r2.rid))

  rids = array([r.rid for r in rlist], int32)
  h5file.createArray(root, 'region_ids', rids)

  rcolors = array([r.color for r in rlist], float32)
  h5file.createArray(root, 'region_colors', rcolors)

  refpts = array([r.max_point for r in rlist], float32)
  h5file.createArray(root, 'ref_points', refpts)

  slev = array([r.smoothing_level for r in rlist], float32)
  h5file.createArray(root, 'smoothing_levels', slev)

  pids = array([(r.preg.rid if r.preg else 0) for r in rlist], int32)
  h5file.createArray(root, 'parent_ids', pids)

  map = seg.seg_map
  if map:
    d = map.data
    a.map_path = d.path
    a.map_size = array(d.size, int32)

  if not seg.map_level is None:
    a.map_level = seg.map_level

  t = seg.point_transform()
  if t:
    from numpy import array, float32
    a.ijk_to_xyz_transform = array(t, float32)

  if seg.adj_graph:
    write_skeleton(h5file, seg.adj_graph)
                          
  h5file.close()
  seg.path = path

# -----------------------------------------------------------------------------
#
def show_save_dialog(seg):

  def save ( okay, dialog ):
    if okay:
      paths = dialog.getPaths ( )
      if paths:
        write_segmentation ( seg, paths[0] )

  if hasattr(seg, 'path'):
    import os.path
    idir, ifile = os.path.split(seg.path)
  else:
    idir = None
    ifile = seg.name
                      
  from OpenSave import SaveModeless
  SaveModeless ( title = 'Save Segmentation %s' % seg.name,
                 initialdir = idir, initialfile = ifile, command = save )

# -----------------------------------------------------------------------------
#
def write_skeleton(h5file, mset):

  import StringIO
  s = StringIO.StringIO()
  mset.save_as_xml(s)
  import numpy
  skel = numpy.char.array(str(s.getvalue()), itemsize = 1)
  s.close()
  h5file.createArray(h5file.root, 'skeleton', skel)

# -----------------------------------------------------------------------------
#
def read_segmentation(path, open = True, task = None):
    
  import tables
  f = tables.openFile(path)

  r = f.root
  a = r._v_attrs
  for n in ('format', 'format_version', 'name'):
    if not n in a:
      raise ValueError, 'Segmentation file does not have "%s" attribute' % n
  if a.format != 'segger':
    raise ValueError, 'Segmentation file format is not "segger"'
  if a.format_version != 2:
    raise ValueError, 'Segmentation file format is not 2'

  from regions import Segmentation
  s = Segmentation(a.name, open = open)

  if 'map_path' in a:
    s.map_path = a.map_path
  if 'map_level' in a:
    s.map_level = a.map_level

  if 'ijk_to_xyz_transform' in a:
    s.ijk_to_xyz_transform = a.ijk_to_xyz_transform

  s.mask = r.mask.read()
  rids = r.region_ids.read()
  rcolors = r.region_colors.read()
  refpts = r.ref_points.read()
  slevels = r.smoothing_levels.read() if hasattr(r, 'smoothing_levels') else None
  pids = r.parent_ids.read()


  from chimera import CancelOperation
  try:
    create_regions(s, rids, rcolors, refpts, slevels, pids, task)
  except CancelOperation:
    f.close()
    raise
  
  read_skeleton(f, s)
  
  f.close()

  s.path = path

  return s


# -----------------------------------------------------------------------------
#
def create_regions(s, rids, rcolors, refpts, slevels, pids, task):

  if task:
    task.updateStatus('Making ID table')
  id_to_index = dict([(id,i) for i,id in enumerate(rids)])

  if task:
    task.updateStatus('Computing region points')
  id_to_points = {}
  import _segment
  rpoints = _segment.region_index_lists(s.mask)
  for rp in rpoints:
    i,j,k = rp[0]
    rid = s.mask[k,j,i]
    id_to_points[rid] = rp

  if task:
    task.updateStatus('Collecting child region IDs')
  id_to_child_ids = {}
  n = len(rids)
  for i in range(n):
    pid = pids[i]
    if pid > 0:
      if pid in id_to_child_ids:
        id_to_child_ids[pid].append(rids[i])
      else:
        id_to_child_ids[pid] = [rids[i]]

  if task:
    task.updateStatus('Ordering IDs')
  from regions import Region
  ids = depth_order(rids, id_to_child_ids, set())
  rlist = []
  for c,rid in enumerate(ids):
    if rid in id_to_child_ids:
      children = [s.id_to_region[cid] for cid in id_to_child_ids[rid]]
    else:
      children = []
    points = id_to_points.get(rid, None)
    i = id_to_index[rid]
    r = Region(s, rid, points, refpts[i], children)
    # TODO: Get wrappy error setting surface piece color to numpy array.
    r.color = tuple(rcolors[i])
    if not slevels is None:
      r.smoothing_level = slevels[i]
    rlist.append(r)
    if task and c % 1000 == 0:
      task.updateStatus('Created %d of %d regions' % (c,n))

  if not slevels is None:
    s.smoothing_level = max(slevels)

  return rlist

# -----------------------------------------------------------------------------
#
def depth_order(rids, id_to_child_ids, used):

  idlist = []
  for rid in rids:
    if not rid in used:
      used.add(rid)
      if rid in id_to_child_ids:
        cids = id_to_child_ids[rid]
        idlist.extend(depth_order(cids, id_to_child_ids, used))
      idlist.append(rid)
  return idlist

# -----------------------------------------------------------------------------
#
def show_open_dialog(dir, callback):

  def open ( okay, dialog ):
    if okay:
      paths_types = dialog.getPathsAndTypes ( )
      if paths_types:
        callback ( paths_types )
                      
  from OpenSave import OpenModeless
  OpenModeless ( title = 'Open Segmentation',
                 initialdir = dir,
                 filters = [('Segmentation', ['*.seg']),
                            ('Old regions file', ['*_regions'])],
                 defaultFilter = 'Segmentation',
                 command = open )

# -----------------------------------------------------------------------------
#
def read_skeleton(f, s):

  a = f.root
  if not 'skeleton' in a:
    return

  sks = a.skeleton.read().tostring()
  import StringIO
  xml = StringIO.StringIO(sks)
  from VolumePath import markerset
  marker_sets = markerset.load_marker_set_xml(xml, model_id = s.id)
  skel = marker_sets[0]
  skel.show_model(True)

  # Map markers to regions
  id2r = dict([(r.rid, r) for r in s.all_regions()])
  for m in skel.markers():
    rid = int(m.extra_attributes['region_id'])
    m.region = id2r.get(rid, None)
    if m.region is None:
      print 'missing skeleton region', rid

  s.adj_graph = skel
